"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/lib/constraint.py
"""

import abc
import collections
import logging
import re
from copy import deepcopy
from itertools import chain

from spike.romancal.associations_pool import PoolRow
from spike.romancal.associations_lib_process_list import ListCategory, ProcessList
from spike.romancal.associations_lib_utilities import evaluate, getattr_from_list, is_iterable

__all__ = [
    "AttrConstraint",
    "Constraint",
    "ConstraintTrue",
    "SimpleConstraint",
]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SimpleConstraintABC(abc.ABC):
    """Simple Constraint ABC

    Parameters
    ----------
    init : dict
        dict where the key:value pairs define
        the following parameters

    value : object or None
        Value that must be matched.

    name : str or None
        Option name for constraint

    **kwargs : key:value pairs
        Other initialization parameters

    Attributes
    ----------
    matched : bool
        Last call to `check_and_set`
    """

    # Attributes to show in the string representation.
    _str_attrs = ("name", "value")

    def __init__(self, init=None, value=None, name=None, **kwargs):
        # Defined attributes
        self.value = value
        self.name = name
        self.matched = False

        if init is not None:
            self.__dict__.update(init)
        else:
            self.__dict__.update(kwargs)

    @abc.abstractmethod
    def check_and_set(self, item):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        self.matched = True
        return self.matched, []

    @property
    def dup_names(
        self,
    ):  #  -> dict[str, list[typing.Union[SimpleConstraint, Constraint]]]
        """Return dictionary of constraints with duplicate names

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Returns
        -------
        dups : {str: [constraint[,...]][,...]}
            Returns a mapping between the duplicated name
            and all the constraints that define that name.
        """
        return {}

    @property
    def id(self):
        """Return identifyer for the constraint

        Returns
        -------
        id : str
            The identifyer
        """
        return f"{self.__class__.__name__}:{self.name}"

    def copy(self):
        """Copy ourselves"""
        return deepcopy(self)

    def get_all_attr(
        self, attribute: str
    ):  # -> list[tuple[SimpleConstraint, typing.Any]]:
        """Return the specified attribute

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Parameters
        ----------
        attribute : str
            The attribute to retrieve

        Returns
        -------
        [(self, value)] : [(SimpleConstraint, object)]
            The value of the attribute in a tuple. If there is no attribute,
            an empty tuple is returned.
        """
        value = getattr(self, attribute)
        if value is not None:
            return [(self, value)]
        return []

    # Make iterable to work with `Constraint`.
    # Since this is a leaf, simple return ourselves.
    def __iter__(self):
        yield self

    def __repr__(self):
        result = f"{self.__class__.__name__}({str(self.__dict__)})"
        return result

    def __str__(self):
        attrs = {str_attr: getattr(self, str_attr) for str_attr in self._str_attrs}
        return f"{self.__class__.__name__}({attrs})"


class ConstraintTrue(SimpleConstraintABC):
    """Always return True"""

    def check_and_set(self, item):
        return super().check_and_set(item)


class SimpleConstraint(SimpleConstraintABC):
    """A basic constraint

    Parameters
    ----------
    init : dict
        dict where the key:value pairs define
        the following parameters

    value : object or None
        Value that must be matched.
        If None, any retrieved value will match.

    sources : func(item) or None
        Function taking `item` as argument used to
        retrieve a value to check against.
        If None, the item itself is used as the value.

    force_unique : bool
        If the constraint is satisfied, reset `value`
        to the value of the source.

    test : function
        The test function for the constraint.
        Takes two arguments:

            - constraint
            - object to compare against.

        Returns a boolean.
        Default is `SimpleConstraint.eq`

    name : str or None
        Option name for constraint

    reprocess_on_match : bool
        Reprocess the item if the constraint is satisfied.

    reprocess_on_fail : bool
        Reprocess the item if the constraint is not satisfied.

    work_over : ListCategory.[BOTH, EXISTING, RULES]
        The condition on which this constraint should operate.

    reprocess_rules : [rule[,..]] or None
        List of rules to be applied to.
        If None, calling function will determine the ruleset.
        If empty, [], all rules will be used.

    Attributes
    ----------
    All `Parameters` are also `Attributes`

    Examples
    --------

    Create a constraint where the attribute `attr` of an object
    matches the value `my_value`:

    >>> c = SimpleConstraint(value='my_value')
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    To check a constraint, call `check_and_set`. A successful match
    will return a tuple of `True` and a reprocess list.
    >>> item = 'my_value'
    >>> c.check_and_set(item)
    (True, [])

    If it doesn't match, `False` will be returned.
    >>> bad_item = 'not_my_value'
    >>> c.check_and_set(bad_item)
    (False, [])

    A `SimpleConstraint` can also be initialized by a `dict`
    of the relevant parameters:
    >>> init = {'value': 'my_value'}
    >>> c = SimpleConstraint(init)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    If the value to check is `None`, the `SimpleConstraint` will
    successfully match whatever object given. However, a new `SimpleConstraint`
    will be returned where the `value` is now set to whatever the attribute
    was of the object.
    >>> c = SimpleConstraint(value=None)
    >>> matched, reprocess = c.check_and_set(item)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    This behavior can be overridden by the `force_unique` parameter:
    >>> c = SimpleConstraint(value=None, force_unique=False)
    >>> matched, reprocess = c.check_and_set(item)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': None})

    """

    def __init__(
        self,
        init=None,
        sources=None,
        force_unique=True,
        test=None,
        reprocess_on_match=False,
        reprocess_on_fail=False,
        work_over=ListCategory.BOTH,
        reprocess_rules=None,
        **kwargs,
    ):
        # Defined attributes
        self.sources = sources
        self.force_unique = force_unique
        self.test = test
        self.reprocess_on_match = reprocess_on_match
        self.reprocess_on_fail = reprocess_on_fail
        self.work_over = work_over
        self.reprocess_rules = reprocess_rules
        super().__init__(init=init, **kwargs)

        # Give defaults some real meaning.
        if self.sources is None:
            self.sources = lambda item: item
        if test is None:
            self.test = self.eq

    def check_and_set(self, item):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        source_value = self.sources(item)

        satisfied = True
        if self.value is not None:
            satisfied = self.test(self.value, source_value)
        self.matched = satisfied

        if self.matched:
            if self.force_unique:
                self.value = source_value

        # Determine reprocessing
        reprocess = []
        if (self.matched and self.reprocess_on_match) or (
            not self.matched and self.reprocess_on_fail
        ):
            reprocess.append(
                ProcessList(
                    items=[item],
                    work_over=self.work_over,
                    rules=self.reprocess_rules,
                    trigger_constraints=[self.id],
                )
            )

        return self.matched, reprocess

    def eq(self, value1, value2):
        """True if constraint.value and item are equal."""
        return value1 == value2


class AttrConstraint(SimpleConstraintABC):
    """Test attribute of an item

    Parameters
    ----------
    sources : [str[,...]]
        List of attributes to query

    value : str, function or None
        The value to check for. If None and
        `force_unique`, any value in the first
        available source will become the value.
        If function, the function takes no arguments
        and returns a string.

    evaluate : bool
        Evaluate the item's value before checking condition.

    force_reprocess : ListCategory.state or False
        Add item back onto the reprocess list using
        the specified `~jwst.associations.ProcessList` work over state.

    force_unique : bool
        If the initial value is `None` or a list of possible values,
        the constraint will be modified to be the value first matched.

    invalid_values : [str[,...]]
        List of values that are invalid in an item.
        Will cause a non-match.

    name : str or None
        Name of the constraint.

    only_on_match : bool
        If `force_reprocess`, only do the reprocess
        if the entire constraint is satisfied.

    onlyif : function
        Boolean function that takes `item` as argument.
        If True, the rest of the condition is checked. Otherwise
        return as a matched condition

    required : bool
        One of the sources must exist. Otherwise,
        return as a matched constraint.

    Attributes
    ----------
    found_values : set(str[,...])
        Set of actual found values for this condition.

    matched : bool
        Last result of `check_and_set`

    """

    # Attributes to show in the string representation.
    _str_attrs = ("name", "sources", "value")

    def __init__(
        self,
        init=None,
        sources=None,
        evaluate=False,
        force_reprocess=False,
        force_undefined=False,
        force_unique=True,
        invalid_values=None,
        only_on_match=False,
        onlyif=None,
        required=True,
        **kwargs,
    ):
        # Attributes
        self.sources = sources
        self.evaluate = evaluate
        self.force_reprocess = force_reprocess
        self.force_undefined = force_undefined
        self.force_unique = force_unique
        self.invalid_values = invalid_values
        self.only_on_match = only_on_match
        self.onlyif = onlyif
        self.required = required
        super().__init__(init=init, **kwargs)

        # Give some defaults real meaning.
        if invalid_values is None:
            self.invalid_values = []
        if onlyif is None:
            self.onlyif = lambda item: True

        # Haven't actually matched anything yet.
        self.found_values = set()
        self.matched = False

    def check_and_set(self, item):
        """Check and set constraints based on item

        Parameters
        ----------
        item : dict
            The item to check on.

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        reprocess = []

        # Only perform check on specified `onlyif` condition
        if not self.onlyif(item):
            if self.force_reprocess:
                reprocess.append(
                    ProcessList(
                        items=[item],
                        work_over=self.force_reprocess,
                        only_on_match=self.only_on_match,
                        trigger_constraints=[self.id],
                    )
                )
            self.matched = True
            return self.matched, reprocess

        # Get the condition information.
        try:
            source, value = getattr_from_list(
                item, self.sources, invalid_values=self.invalid_values
            )
        except KeyError:
            if self.required and not self.force_undefined:
                self.matched = False
                return self.matched, reprocess
            else:
                self.matched = True
                return self.matched, reprocess
        else:
            if self.force_undefined:
                self.matched = False
                return self.matched, reprocess

        evaled = value
        if self.evaluate:
            evaled = evaluate(value)

        # If the constraint has no value to check against, and given
        # value evaluates to a list, the item must be duplicated,
        # with each value from its list, and all the new items reprocessed.
        # Otherwise, the value is the value to set the constraint by.
        if self.value is None:
            if is_iterable(evaled):
                reprocess.append(reprocess_multivalue(item, source, evaled, self))
                self.matched = False
                return self.matched, reprocess
            value = str(evaled)

        # Else, the constraint does have a value. Check against it.
        else:
            if callable(self.value):
                match_value = self.value()
            else:
                match_value = self.value
            if not is_iterable(evaled):
                evaled = [evaled]
            matched = []
            for evaled_item in evaled:
                value = str(evaled_item)
                if meets_conditions(value, match_value):
                    matched.append(evaled_item)

            # If the condition is not matched, leave now.
            if not matched:
                self.matched = False
                return self.matched, reprocess

            # Matches were found. The first match becomes
            # the value for the current constraint.
            # The rest get reprocessed.
            value = str(matched[0])
            if matched[1:]:
                reprocess.append(reprocess_multivalue(item, source, matched[1:], self))

        # At this point, the constraint has passed.
        # Fix the conditions.
        escaped_value = re.escape(value)
        self.found_values.add(escaped_value)
        if self.force_unique:
            self.value = escaped_value
            self.sources = [source]
            self.force_unique = False

        # If required to reprocess, add to the reprocess list.
        if self.force_reprocess:
            reprocess.append(
                ProcessList(
                    items=[item],
                    work_over=self.force_reprocess,
                    only_on_match=self.only_on_match,
                    trigger_constraints=[self.id],
                )
            )

        # That's all folks
        self.matched = True
        return self.matched, reprocess


class Constraint:
    """Constraint that is made up of SimpleConstraints

    Parameters
    ----------
    init : object or [object[,...]]
        A single object or list of objects where the
        objects are as follows.
        - SimpleConstraint or subclass
        - Constraint

    reduce : function
        A reduction function with signature `x(iterable)`
        where `iterable` is the `components` list. Returns
        boolean indicating state of the components.
        Default value is `Constraint.all`

    name : str or None
        Optional name for constraint.

    reprocess_on_match : bool
        Reprocess the item if the constraint is satisfied.

    reprocess_on_fail : bool
        Reprocess the item if the constraint is not satisfied.

    work_over : ListCategory.[BOTH, EXISTING, RULES]
        The condition on which this constraint should operate.

    reprocess_rules : [rule[,..]] or None
        List of rules to be applied to.
        If None, calling function will determine the ruleset.
        If empty, [], all rules will be used.

    Attributes
    ----------
    constraints : [Constraint[,...]]
        List of `Constraint` or `SimpleConstraint` that
        make this constraint.

    matched : bool
        Result of the last `check_and_set`

    reduce : function
        A reduction function with signature `x(iterable)`
        where `iterable` is the `components` list. Returns
        boolean indicating state of the components.
        Predefined functions are:
        - `all`: True if all components return True
        - `any`: True if any component returns True

    Notes
    -----
    Named constraints can be accessed directly through indexing:

    >>> c = Constraint(SimpleConstraint(name='simple', value='a_value'))
    >>> c['simple']  # doctest: +SKIP
    SimpleConstraint({'sources': <function SimpleConstraint.__init__.<locals>.<lambda> at 0x7f8be05f5730>,
                      'force_unique': True,
                      'test': <bound method SimpleConstraint.eq of SimpleConstraint({...})>,
                      'reprocess_on_match': False,
                      'reprocess_on_fail': False,
                      'work_over': 1,
                      'reprocess_rules': None,
                      'value': 'a_value',
                      'name': 'simple',
                      'matched': False})
    """  # noqa: E501

    def __init__(
        self,
        init=None,
        reduce=None,
        name=None,
        reprocess_on_match=False,
        reprocess_on_fail=False,
        work_over=ListCategory.BOTH,
        reprocess_rules=None,
    ):
        self.constraints = []

        # Initialize from named parameters
        self.reduce = reduce
        self.name = name
        self.reprocess_on_match = reprocess_on_match
        self.reprocess_on_fail = reprocess_on_fail
        self.work_over = work_over
        self.reprocess_rules = reprocess_rules

        # Initialize from a structure.
        if init is None:
            pass
        elif isinstance(init, list):
            self.constraints = init
        elif isinstance(init, Constraint):
            self.reduce = init.reduce
            self.name = init.name
            self.reprocess_on_match = init.reprocess_on_match
            self.reprocess_on_fail = init.reprocess_on_fail
            self.work_over = init.work_over
            self.reprocess_rules = init.reprocess_rules
            self.constraints = deepcopy(init.constraints)
        elif isinstance(init, SimpleConstraintABC):
            self.constraints = [init]
        else:
            raise TypeError(
                "Invalid initialization value type {}.\nValid types are"
                " `SimpleConstraint`, `Constraint`,\nor subclass.".format(type(init))
            )

        # Give some defaults real meaning.
        self.matched = False
        if self.reduce is None:
            self.reduce = self.all

    @property
    def dup_names(
        self,
    ):  # -> dict[str, list[typing.Union[SimpleConstraint, Constraint]]]:
        """Return dictionary of constraints with duplicate names

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Returns
        -------
        dups : {str: [constraint[,...]][,...]}
            Returns a mapping between the duplicated name
            and all the constraints that define that name.
        """
        attrs = self.get_all_attr("name")
        constraints, names = zip(*attrs)
        dups = [name for name, count in collections.Counter(names).items() if count > 1]
        result = collections.defaultdict(list)
        for name, constraint in zip(names, constraints):
            if name in dups:
                result[name].append(constraint)

        # Turn off the defaultdict factory.
        result.default_factory = None
        return result

    @property
    def id(self):
        """Return identifyer for the constraint

        Returns
        -------
        id : str
            The identifyer
        """
        return f"{self.__class__.__name__}:{self.name}"

    def append(self, constraint):
        """Append a new constraint"""
        self.constraints.append(constraint)

    def check_and_set(self, item, work_over=ListCategory.BOTH):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - success : True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        if work_over not in (self.work_over, ListCategory.BOTH):
            return False, []

        # Do we have positive?
        self.matched, reprocess = self.reduce(item, self.constraints)

        # Determine reprocessing
        if (self.matched and self.reprocess_on_match) or (
            not self.matched and self.reprocess_on_fail
        ):
            reprocess.append(
                [
                    ProcessList(
                        items=[item],
                        work_over=self.work_over,
                        rules=self.reprocess_rules,
                        trigger_constraints=[self.id],
                    )
                ]
            )

        return self.matched, list(chain(*reprocess))

    def copy(self):
        """Copy ourselves"""
        return deepcopy(self)

    def get_all_attr(
        self, attribute: str
    ):  # -> list[tuple[typing.Union[SimpleConstraint, Constraint], typing.Any]]:
        """Return the specified attribute

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Parameters
        ----------
        attribute : str
            The attribute to retrieve

        Returns
        -------
        result : [(SimpleConstraint or Constraint, object)[,...]]
            The list of values of the attribute in a tuple. If there is no attribute,
            an empty tuple is returned.

        Raises
        ------
        AttributeError
            If the attribute is not found.
        """
        result = []
        value = getattr(self, attribute)
        if value is not None:
            result = [(self, value)]
        for constraint in self.constraints:
            result.extend(constraint.get_all_attr(attribute))

        return result

    @staticmethod
    def all(item, constraints):
        """Return positive only if all results are positive."""

        # If there are no constraints, there is nothing to match.
        # Result is false.
        if len(constraints) == 0:
            return False, []

        # Find all negatives. Note first negative
        # that requires reprocessing and how many
        # negatives do not.
        all_match = True
        negative_reprocess = None
        to_reprocess = []
        for constraint in constraints:
            match, reprocess = constraint.check_and_set(item)

            if match:
                if all_match:
                    to_reprocess.append(reprocess)
            else:
                all_match = False

                # If not match and no reprocessing, then fail
                # completely. However, if there is reprocessing, take
                # the first one. Continue to check to ensure
                # there is no further complete fail.
                if len(reprocess) == 0:
                    negative_reprocess = None
                    break
                elif negative_reprocess is None:
                    negative_reprocess = [reprocess]

        if not all_match:
            if negative_reprocess is not None:
                to_reprocess = negative_reprocess
            else:
                to_reprocess = []

        return all_match, to_reprocess

    @staticmethod
    def any(item, constraints):
        """Return the first successful constraint."""
        # If there are no constraints, there is nothing to match.
        # Result is false.
        if len(constraints) == 0:
            return False, []

        to_reprocess = []
        for constraint in constraints:
            match, reprocess = constraint.check_and_set(item)
            if match:
                to_reprocess = [reprocess]
                break
            to_reprocess.append(reprocess)
        return match, to_reprocess

    @staticmethod
    def notany(item, constraints):
        """True if none of the constraints match"""
        match, to_reprocess = Constraint.any(item, constraints)
        return not match, to_reprocess

    @staticmethod
    def notall(item, constraints):
        """True if not all of the constraints match"""
        match, to_reprocess = Constraint.all(item, constraints)
        return not match, to_reprocess

    def __delitem__(self, key):
        """Not implemented"""
        raise NotImplementedError("Cannot delete a constraint by index.")

    # Make iterable
    def __iter__(self):
        yield from chain(*map(iter, self.constraints))

    # Index implementation
    def __getitem__(self, key):
        """Retrieve a named constraint"""
        for constraint in self.constraints:
            name = getattr(constraint, "name", None)
            if name is not None and name == key:
                return constraint
            try:
                found = constraint[key]
            except (KeyError, TypeError):
                pass
            else:
                return found
        raise KeyError(f"Constraint {key} not found")

    def __repr__(self):
        result = "{}(name={}).{}([{}])".format(
            self.__class__.__name__,
            str(getattr(self, "name", None)),
            str(self.reduce.__name__),
            "".join([repr(constraint) for constraint in self.constraints]),
        )
        return result

    def __setitem__(self, key, value):
        """Not implemented"""
        raise NotImplementedError("Cannot set constraints by index.")

    def __str__(self):
        result = "\n".join(
            [str(constraint) for constraint in self if constraint.name is not None]
        )
        return result


# ---------
# Utilities
# ---------
def meets_conditions(value, conditions):
    """Check whether value meets any of the provided conditions

    Parameters
    ----------
    values : str
        The value to be check with.

    condition : regex,
        Regular expressions to match against.

    Returns
    -------
    True if any condition is meant.
    """

    if not is_iterable(conditions):
        conditions = [conditions]
    for condition in conditions:
        condition = "".join(["^", condition, "$"])
        match = re.match(condition, value, flags=re.IGNORECASE)
        if match:
            return True
    return False


def reprocess_multivalue(item, source, values, constraint):
    """Reprocess items that have a list of values

    Parameters
    ----------
    item : dict
        The item.

    source : str
        The attribute which has the multi-values.

    values : list
        The list of values

    constraint : Constraint
        The constraint which is triggering the reprocessing.

    Returns
    -------
    process_list : ProcessList
        The process list to put on the reprocess queue
    """
    reprocess_items = []
    for value in values:
        new_item = PoolRow(item)
        new_item[source] = str(value)
        reprocess_items.append(new_item)
    process_list = ProcessList(
        items=reprocess_items, trigger_constraints=[constraint.id]
    )
    return process_list


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/config_parser.py
"""

import logging
import os
import os.path
import textwrap
import warnings
from inspect import isclass

from asdf import ValidationError as AsdfValidationError
from asdf import open as asdf_open
from astropy.extern.configobj.configobj import (
    ConfigObj,
    Section,
    flatten_errors,
    get_extra_values,
)
from astropy.extern.configobj.validate import ValidateError, Validator, VdtTypeError

from spike.stpipe import utilities
from spike.stpipe.config import StepConfig
from spike.stpipe.datamodel import AbstractDataModel
from spike.stpipe.utilities import _not_set

# Configure logger
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ValidationError(Exception):
    pass


def _get_input_file_check(root_dir):
    from spike.stpipe import cmdline

    root_dir = root_dir or ""

    def _input_file_check(path):
        if not isinstance(path, cmdline.FromCommandLine):
            try:
                path = str(path)
            except ValueError:
                pass

            path = os.path.join(root_dir, path)

        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValidateError(f"Path {path!r} does not exist")

        return path

    return _input_file_check


def _get_output_file_check(root_dir):
    from spike.stpipe import cmdline

    root_dir = root_dir or ""

    def _output_file_check(path):
        if not isinstance(path, cmdline.FromCommandLine):
            try:
                path = str(path)
            except ValueError:
                pass

            path = os.path.join(root_dir, path)

        path = os.path.abspath(path)
        dir_ = os.path.dirname(path)
        if dir_ and not os.path.exists(dir_):
            os.makedirs(dir_)

        return path

    return _output_file_check


def _is_datamodel(value):
    """Verify that value is a DataModel."""
    if isinstance(value, AbstractDataModel):
        return value

    raise VdtTypeError(value)


def _is_string_or_datamodel(value):
    """Verify that value is either a string (nominally a reference file path)
    or a DataModel (possibly one with no corresponding file.)
    """
    if isinstance(value, AbstractDataModel):
        return value

    if isinstance(value, str):
        return value

    raise VdtTypeError(value)


def load_config_file(config_file):
    """
    Read the file `config_file` and return the parsed configuration.
    """
    if not os.path.isfile(config_file):
        raise ValueError(f"Config file {config_file} not found.")
    try:
        with asdf_open(config_file) as asdf_file:
            return _config_obj_from_asdf(asdf_file)
    except (AsdfValidationError, ValueError):
        logger.debug(
            "Config file did not parse as ASDF. Trying as ConfigObj: %s", config_file
        )
        return ConfigObj(config_file, raise_errors=True)


def _config_obj_from_asdf(asdf_file):
    config = StepConfig.from_asdf(asdf_file)
    return _config_obj_from_step_config(config)


def _config_obj_from_step_config(config):
    configobj = ConfigObj()
    merge_config(configobj, config.parameters)
    merge_config(configobj, {"class": config.class_name, "name": config.name})
    if len(config.steps) > 0:
        merge_config(
            configobj,
            {"steps": {s.name: _config_obj_from_step_config(s) for s in config.steps}},
        )
    return configobj


def get_merged_spec_file(cls, preserve_comments=_not_set):
    """
    Creates a merged spec file for a Step class and all of its
    subclasses.

    Parameters
    ----------
    cls: `Step`-derived or `Step` instance
        A class or instance of a `Step`-based class.

    preserve_comments : bool, optional
        This argument is deprecated and appears to have no
        effect on the returned spec.
        When True, preserve the comments in the spec file
    """
    if not isclass(cls):
        cls = cls.__class__
    subclasses = cls.mro()
    subclasses.reverse()

    config = ConfigObj()
    cfg = None
    for subclass in subclasses:
        cfg = load_spec_file(subclass, preserve_comments=preserve_comments)
        if cfg:
            merge_config(config, cfg)

    if cfg is not None:
        config.initial_comment = cfg.initial_comment
        config.final_comment = cfg.final_comment

    return config


def load_spec_file(cls, preserve_comments=_not_set):
    """
    Load the spec file corresponding to the given class.

    Parameters
    ----------
    cls: `Step`-derived or `Step` instance
        A class or instance of a `Step`-based class.

    preserve_comments: bool
        This argument is deprecated and appears to have no
        effect on the returned spec.
        True to keep comments in the resulting `ConfigObj`

    Returns
    -------
    spec_file: ConfigObj
        The resulting configuration object
    """
    if preserve_comments is not _not_set:
        msg = "preserve_comments is deprecated"
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
    # Don't use 'hasattr' here, because we don't want to inherit spec
    # from the base class.
    if not isclass(cls):
        cls = cls.__class__
    if "spec" in cls.__dict__:
        spec = cls.spec.strip()
        spec_file = textwrap.dedent(spec)
        spec_file = spec_file.split("\n")
        encoded = []
        for line in spec_file:
            if isinstance(line, str):
                encoded.append(line.encode("utf8"))
            else:
                encoded.append(line)
        spec_file = encoded
    else:
        spec_file = utilities.find_spec_file(cls)
    if spec_file:
        return ConfigObj(
            spec_file,
            raise_errors=True,
            list_values=False,
        )
    return None


def merge_config(into, new):
    """
    Merges a configuration tree into another one.

    Unlike merge in configobj itself, this one updates inline
    comments and does not treat non-dict Mappings as config
    sections.

    Parameters
    ----------
    into : `configobj.Section`
        The configuration tree to merge into

    new : `configobj.Section` or dict
        The source of new configuration values
    """
    if isinstance(new, Section):
        defaults = new.defaults
        inline_comments = new.inline_comments
        comments = new.comments
    else:
        defaults = set()
        inline_comments = {}
        comments = {}

    for key, val in new.items():
        if isinstance(val, Section):
            if key not in into:
                section = Section(
                    into,
                    into.depth + 1,
                    into.main,
                    name=key,
                )
                into[key] = section
            merge_config(into[key], val)
        elif key not in defaults:
            # The unrepr flag is configured to only allow dicts to be
            # treated as config sections.  This prevents other
            # Mappings (such as DataModel, used in reference overrides)
            # from being converted.
            into.__setitem__(key, val, unrepr=not isinstance(val, dict))
            into.inline_comments[key] = inline_comments.get(key)
            into.comments[key] = comments.get(key)


def config_from_dict(d, spec=None, root_dir=None, allow_missing=False):
    """
    Create a ConfigObj from a dict.

    Parameters
    ----------
    d: dict
        The dictionary to merge into the resulting ConfigObj.

    spec: ConfigObj
        The specification to validate against.
        If None, just convert dictionary into a ConfigObj.

    root_dir: str
        The base directory to use for file-based parameters.

    allow_missing: bool
        If a parameter is not defined and has no default in the spec,
        set that parameter to its specification.
    """
    config = ConfigObj()

    merge_config(config, d)

    if spec:
        validate(config, spec, root_dir=root_dir, allow_missing=allow_missing)
    else:
        config.walk(string_to_python_type)

    return config


def validate(
    config, spec, section=None, validator=None, root_dir=None, allow_missing=False
):
    """
    Parse config_file, in INI format, and do validation with the
    provided specfile.

    Parameters
    ----------
    config: ConfigObj
        The configuration to validate.

    spec: ConfigObj
        The specification to validate against.

    section: ConfigObj or None
        The specific section of config to validate.
        If None, then all sections are validated.

    validator: extern.configobj.validator.Validator or None
        The validator to use. If None, the default will be used.

    root_dir: str
        The directory to use as the basis for any file-based parameters.

    allow_missing: bool
        If a parameter is not defined and has no default in the spec,
        set that parameter to its specification.
    """
    if spec is None:
        config.walk(string_to_python_type)
        return config

    if validator is None:
        validator = Validator()
        validator.functions["input_file"] = _get_input_file_check(root_dir)
        validator.functions["output_file"] = _get_output_file_check(root_dir)
        validator.functions["is_datamodel"] = _is_datamodel
        validator.functions["is_string_or_datamodel"] = _is_string_or_datamodel

    orig_configspec = config.main.configspec
    config.main.configspec = spec

    try:
        if config.main != config:
            section = config
        else:
            section = None

        errors = config.main.validate(
            validator,
            preserve_errors=True,
            section=section,
        )

        messages = []
        if errors is not True:
            for section_list, key, err in flatten_errors(config, errors):
                if key is not None:
                    section_list.append(key)
                else:
                    section_list.append("[missing section]")
                section_string = "/".join(section_list)
                if err is False:
                    if allow_missing:
                        config[key] = spec[key]
                        continue

                    err = "missing"

                messages.append(f"Config parameter {section_string!r}: {err}")

        extra_values = get_extra_values(config)
        if extra_values:
            sections, name = extra_values[0]
            if len(sections) == 0:
                sections = "root"
            else:
                sections = "/".join(sections)
            messages.append(f"Extra value {name!r} in {sections}")

        if len(messages):
            raise ValidationError("\n".join(messages))
    finally:
        config.main.configspec = orig_configspec

    return config


def string_to_python_type(section, key):
    """
    Do blind type inferring.
    """
    # We parse scalars and lists.
    val = section[key]
    if isinstance(val, list):
        typed_val = [_parse(x) for x in val]
    else:
        typed_val = _parse(val)
    section[key] = typed_val


def _parse(val):
    """
    Parse scalar strings into scalar python types.
    """
    if val.lower() == "true":
        return True

    if val.lower() == "false":
        return False

    try:
        return int(val)
    except ValueError:
        pass

    try:
        return float(val)
    except ValueError:
        pass
    return str(val)


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/lib/signal_slot.py
"""


from collections import namedtuple
import inspect
import logging


__all__ = ['Signal',
           'Signals',
           'SignalsNotAClass']

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

"""Slot data structure
"""
Slot = namedtuple('Slot', ['func', 'single_shot'])


class Signal():
    """Signal

    A Signal, when triggered, call the connected slots.

    Parameters
    ----------
    funcs: func[, ...]
        Remaining arguments will be functions to connect
        to this signal.

    Attributes
    ----------
    enabled: bool
        If True, the slots are called. Otherwise, nothing
        happens when triggered.

    """

    def __init__(self, *funcs):
        self._slots = list()
        self._enabled = True
        self._states = list()

        for func in funcs:
            self.connect(func)

    def emit(self, *args, **kwargs):
        """Invoke slots attached to the signal

        No return of results is expected.

        Parameters
        ----------
        args: (arg[, ...])
            Positional arguments to pass to the slots.

        kwargs: {key: value[, ...]}
            Keyword arguments to pass to the slots.
        """
        for _ in self.call(*args, **kwargs):
            pass

    __call__ = emit

    def call(self, *args, **kwargs):
        """Generator returning result of each slot connected.

        Parameters
        ----------
        args: (arg[, ...])
            Positional arguments to pass to the slots.

        kwargs: {key: value[, ...]}
            Keyword arguments to pass to the slots.

        Returns
        -------
        generator
            A generator returning the result from each slot.
        """
        for slot in self.slots:
            try:
                yield slot(*args, **kwargs)
            except Exception as exception:
                logger.debug(
                    'Signal {}: Slot {} raised {}'.format(
                        self.__class__.__name_,
                        slot,
                        exception
                    )
                )

    def reduce(self, *args, **kwargs):
        """Return a reduction of all the slots

        Parameters
        ----------
        args: (arg[, ...])
            Positional arguments to pass to the slots.

        kwargs: {key: value[, ...]}
            Keyword arguments to pass to the slots.

        Returns
        -------
        result: object or (object [,...])
            The result or tuple of results. See `Notes`.


        Notes
        -----

        Each slot is given the results of the previous
        slot as a new positional argument list. As such, if multiple
        arguments are required, each slot should return a tuple that
        can then be passed as arguments to the next function.

        The keyword arguments are simply passed to each slot unchanged.

        There is no guarantee on order which the slots are invoked.

        """
        result = None
        for slot in self.slots:
            result = slot(*args, **kwargs)
            args = result
            if not isinstance(args, tuple):
                args = (args, )
        return result

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, state):
        self.set_enabled(state, push=False)

    def set_enabled(self, state, push=False):
        """Set whether signal is active or not

        Parameters
        ----------
        state: boolean
            New state of signal

        push: boolean
            If True, current state is saved.
        """
        if push:
            self._states.append(self._enabled)
        self._enabled = state

    def reset_enabled(self):
        self._enabled = self._states.pop()

    def connect(self, func, single_shot=False):
        """Connect a function to the signal
        Parameters
        ----------
        func: function or method
            The function/method to call when the signal is activated

        single_shot: bool
            If True, the function/method is removed after being called.
        """
        slot = Slot(
            func=func,
            single_shot=single_shot
        )
        self._slots.append(slot)

    def disconnect(self, func):
        self._slots = [
            slot
            for slot in self._slots
            if slot.func != func
        ]

    def clear(self, single_shot=False):
        """Clear slots

        Parameters
        ----------
        single_shot: bool
            If True, only remove single shot
            slots.
        """
        logger.debug(
            'Signal {}: Clearing slots'.format(
                self.__class__.__name__
            )
        )
        if not single_shot:
            self._slots.clear()
        else:
            self._slots = [
                slot
                for slot in self._slots
                if not slot.single_shot
            ]

    @property
    def slots(self):
        """Generator returning slots"""
        if not self.enabled:
            return

        # No recursive signalling
        self.set_enabled(False, push=True)

        try:
            for slot in self._slots:
                yield slot.func
        finally:
            # Clean out single shots
            self._slots = [
                slot
                for slot in self._slots
                if not slot.single_shot
            ]
            self.reset_enabled()


class SignalsErrorBase(Exception):
    '''Base Signals Error'''

    default_message = ''

    def __init__(self, *args):
        if len(args):
            super(SignalsErrorBase, self).__init__(*args)
        else:
            super(SignalsErrorBase, self).__init__(self.default_message)


class SignalsNotAClass(SignalsErrorBase):
    '''Must add a Signal Class'''
    default_message = 'Signal must be a class.'


class Signals(dict):
    '''Manage the signals.'''

    def __setitem__(self, key, value):
        if key not in self:
            super(Signals, self).__setitem__(key, value)
        else:
            logger.warning('Signals: signal "{}" already exists.'.format(key))

    def __getattr__(self, key):
        for signal in self:
            if signal.__name__ == key:
                return self[signal]
        raise KeyError('{}'.format(key))

    def add(self, signal_class, *args, **kwargs):
        if inspect.isclass(signal_class):
            self.__setitem__(signal_class, signal_class(*args, **kwargs))
        else:
            raise SignalsNotAClass


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/source_catalog/detection.py
"""

"""
Module to detect sources using image segmentation.
"""

import logging
import warnings

from astropy.convolution import Gaussian2DKernel, convolve
from astropy.stats import gaussian_fwhm_to_sigma, SigmaClip
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning
import numpy as np
from photutils.background import Background2D, MedianBackground
from photutils.utils.exceptions import NoDetectionsWarning
from photutils.segmentation import detect_sources, deblend_sources

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class JWSTBackground:
    """
    Class to estimate a 2D background and background RMS noise in an
    image.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The input 2D array.

    box_size : int or array_like (int)
        The box size along each axis.  If ``box_size`` is a scalar then
        a square box of size ``box_size`` will be used.  If ``box_size``
        has two elements, they should be in ``(ny, nx)`` order.

    coverage_mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from calculations. ``coverage_mask``
        should be `True` where there is no coverage (i.e., no data) for
        a given pixel (e.g., blank areas in a mosaic image). It should
        not be used for bad pixels.

    Attributes
    ----------
    background : 2D `~numpy.ndimage`
        The estimated 2D background image.

    background_rms : 2D `~numpy.ndimage`
        The estimated 2D background RMS image.
    """

    def __init__(self, data, box_size=100, coverage_mask=None):
        self.data = data
        self.box_size = np.asarray(box_size).astype(int)  # must be integer
        self.coverage_mask = coverage_mask

    @lazyproperty
    def _background2d(self):
        """
        Estimate the 2D background and background RMS noise in an image.

        Returns
        -------
        background : `photutils.background.Background2D`
            A Background2D object containing the 2D background and
            background RMS noise estimates.
        """
        sigma_clip = SigmaClip(sigma=3.)
        bkg_estimator = MedianBackground()
        filter_size = (3, 3)

        # All data have NaNs.  Suppress warnings about them.
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=AstropyUserWarning)
            try:
                bkg = Background2D(self.data, self.box_size,
                                   filter_size=filter_size,
                                   coverage_mask=self.coverage_mask,
                                   sigma_clip=sigma_clip,
                                   bkg_estimator=bkg_estimator)
            except ValueError:
                # use the entire unmasked array
                bkg = Background2D(self.data, self.data.shape,
                                   filter_size=filter_size,
                                   coverage_mask=self.coverage_mask,
                                   sigma_clip=sigma_clip,
                                   bkg_estimator=bkg_estimator,
                                   exclude_percentile=100.)
                log.info('Background could not be estimated in meshes. '
                         'Using the entire unmasked array for background '
                         f'estimation: bkg_boxsize={self.data.shape}.')

        return bkg

    @lazyproperty
    def background(self):
        """
        The 2D background image.
        """
        return self._background2d.background

    @lazyproperty
    def background_rms(self):
        """
        The 2D background RMS image.
        """
        return self._background2d.background_rms


def make_kernel(kernel_fwhm):
    """
    Make a 2D Gaussian smoothing kernel that is used to filter the image
    before thresholding.

    Filtering the image will smooth the noise and maximize detectability
    of objects with a shape similar to the kernel.

    The kernel must have odd sizes in both X and Y, be centered in the
    central pixel, and normalized to sum to 1.

    Parameters
    ----------
    kernel_fwhm : float
        The full-width at half-maximum (FWHM) of the 2D Gaussian kernel.

    Returns
    -------
    kernel : `astropy.convolution.Kernel2D`
        The output smoothing kernel, normalized such that it sums to 1.
    """
    sigma = kernel_fwhm * gaussian_fwhm_to_sigma
    kernel = Gaussian2DKernel(sigma)
    kernel.normalize(mode='integral')
    return kernel


def convolve_data(data, kernel_fwhm, mask=None):
    """
    Convolve the data with a Gaussian2D kernel.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The 2D array to convolve.

    kernel_fwhm : float
        The full-width at half-maximum (FWHM) of the 2D Gaussian kernel.

    mask : array_like, bool, optional
        A boolean mask with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
    """
    kernel = make_kernel(kernel_fwhm)

    # All data have NaNs.  Supress warnings about them.
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=AstropyUserWarning)
        return convolve(data, kernel, mask=mask, normalize_kernel=True)


class JWSTSourceFinder:
    """
    Class to detect sources, including deblending, using image
    segmentation.

    Parameters
    ----------
    threshold : float
        The data value to be used as the per-pixel detection threshold.

    npixels : int
        The number of connected pixels, each greater than the threshold,
        that an object must have to be detected. ``npixels`` must be a
        positive integer.

    deblend : bool, optional
        Whether to deblend overlapping sources. Source deblending
        requires scikit-image.
    """

    def __init__(self, threshold, npixels, deblend=False):
        self.threshold = threshold
        self.npixels = npixels
        self.deblend = deblend
        self.connectivity = 8
        self.nlevels = 32
        self.contrast = 0.001
        self.mode = 'exponential'

    def __call__(self, convolved_data, mask=None):
        """
        Parameters
        ----------
        convolved_data : 2D `numpy.ndarray`
            The 2D convolved array from which to detect sources.

        mask : array_like, bool, optional
            A boolean mask with the same shape as ``convolved_data``,
            where a `True` value indicates the corresponding element
            of ``convolved_data`` is masked. Masked pixels will not be
            included in any source.

        Returns
        -------
        segment_image : `~photutils.segmentation.SegmentationImage` or `None`
            A 2D segmentation image, with the same shape as the input data,
            where sources are marked by different positive integer values. A
            value of zero is reserved for the background. If no sources are
            found then `None` is returned.
        """
        if mask is not None:
            if mask.all():
                log.error('There are no valid pixels in the image to detect '
                          'sources.')
                return None

        with warnings.catch_warnings():
            # suppress NoDetectionsWarning from photutils
            warnings.filterwarnings('ignore', category=NoDetectionsWarning)

            segment_img = detect_sources(convolved_data, self.threshold,
                                         self.npixels, mask=mask,
                                         connectivity=self.connectivity)
            if segment_img is None:
                log.warning('No sources were found. Source catalog will not '
                            'be created.')
                return None

            # source deblending requires scikit-image
            if self.deblend:
                segment_img = deblend_sources(convolved_data, segment_img,
                                              npixels=self.npixels,
                                              nlevels=self.nlevels,
                                              contrast=self.contrast,
                                              mode=self.mode,
                                              connectivity=self.connectivity,
                                              relabel=True)

        log.info(f'Detected {segment_img.nlabels} sources')
        return segment_img


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/cmdline.py
"""


import io
import os
import os.path
import textwrap

from spike.stpipe import config_parser, log, utilities
from spike.stpipe.step import Step, get_disable_crds_steppars

built_in_configuration_parameters = [
    "debug",
    "logcfg",
    "verbose",
]


def _print_important_message(header, message, no_wrap=None):
    print("-" * 70)
    print(textwrap.fill(header))
    print(
        textwrap.fill(
            message,
            initial_indent="    ",
            subsequent_indent="    ",
        )
    )
    if no_wrap:
        print(no_wrap)
    print("-" * 70)


def _get_config_and_class(identifier):
    """
    Given a file path to a config file or Python module path, return a
    Step class and a configuration object.
    """
    if os.path.exists(identifier):
        config_file = identifier
        config = config_parser.load_config_file(config_file)
        step_class, name = Step._parse_class_and_name(config, config_file=config_file)
    else:
        try:
            step_class = utilities.import_class(
                utilities.resolve_step_class_alias(identifier), Step
            )
        except (ImportError, AttributeError, TypeError) as err:
            raise ValueError(
                f"{identifier!r} is not a path to a config file or a Python Step class"
            ) from err
        # Don't validate yet
        config = config_parser.config_from_dict({})
        name = None
        config_file = None

    return step_class, config, name, config_file


def _build_arg_parser_from_spec(spec, step_class, parent=None):
    """
    Given a configspec, sets up an argparse argument parser that
    understands its arguments.

    The \"path\" in the configspec becomes a dot-separated identifier
    in the commandline arguments.  For example, in the following
    configfile::

        [foo]
          [[bar]]
             baz = 2

    The "baz" variable can be changed with ``--foo.bar.baz=42``.
    """
    import argparse

    # It doesn't translate the configspec types -- it instead
    # will accept any string.  However, the types of the arguments will
    # later be verified by configobj itself.
    parser = argparse.ArgumentParser(
        parents=[parent],
        description=step_class.__doc__,
    )

    def build_from_spec(subspec, parts=None):
        if parts is None:
            parts = []
        for key, val in subspec.items():
            if isinstance(val, dict):
                build_from_spec(val, [*parts, key])
            else:
                comment = subspec.inline_comments.get(key) or ""
                comment = comment.lstrip("#").strip()
                # Only show default value if it is not None or the empty string
                default_value_string = val.split("(")[1].rstrip(")").strip()
                if default_value_string.lstrip("default=") in ["None", "''", '""']:
                    help_string = comment
                else:
                    help_string = f"{comment} [{default_value_string}]"
                argument = "--" + ".".join([*parts, key])
                if argument[2:] in built_in_configuration_parameters:
                    raise ValueError(
                        "The Step's spec is trying to override a built-in parameter"
                        f" {argument!r}"
                    )
                parser.add_argument(
                    "--" + ".".join([*parts, key]),
                    type=str,
                    help=help_string,
                    metavar="",
                )

    build_from_spec(spec)

    parser.add_argument(
        "args",
        nargs="*",
        help="arguments to pass to step",
    )

    return parser


class FromCommandLine(str):
    """
    We need a way to distinguish between config values that come from
    a config file and those that come from the commandline.  For
    example, configfile paths must be resolved against the location of
    the config file.  Commandline paths must be resolved against the
    current working directory.  By setting all commandline overrides
    as instances of this class, we can later (in `config_parser.py`)
    use isinstance to see where the values came from.
    """


def _override_config_from_args(config, args):
    """
    Overrides any configuration values in `config` with values from the
    parsed commandline arguments `args`.
    """

    def set_value(subconf, key, val):
        root, sep, rest = key.partition(".")
        if rest:
            set_value(subconf.setdefault(root, {}), rest, val)
        else:
            val, comment = config._handle_value(val)
            if isinstance(val, str):
                subconf[root] = FromCommandLine(val)
            else:
                subconf[root] = val

    for key, val in vars(args).items():
        if val is not None:
            set_value(config, key, val)


def just_the_step_from_cmdline(args, cls=None):
    """
    Create a step from a configuration file and return it.  Don't run it.

    Parameters
    ----------
    args : list of str
        Commandline arguments

    cls : Step class
        The step class to use.  If none is provided, the step

    Returns
    -------
    step : Step instance
        If the config file has a `class` parameter, or the commandline
        specifies a class, the return value will be as instance of
        that class.

        Any parameters found in the config file or on the commandline
        will be set as member variables on the returned `Step`
        instance.

    step_class: Step class
        As defined by `cls` parameter or .cfg file.

    positional: list of strings
        Positional parameters after arg parsing

    DOES NOT RUN THE STEP
    """
    import argparse

    parser1 = argparse.ArgumentParser(
        description="Run an stpipe Step or Pipeline",
        add_help=False,
    )
    if cls is None:
        parser1.add_argument(
            "cfg_file_or_class",
            type=str,
            nargs=1,
            help="The configuration file or Python class to run",
        )
    else:
        parser1.add_argument(
            "--config-file",
            type=str,
            help="A configuration file to load parameters from",
        )
    parser1.add_argument(
        "--logcfg",
        type=str,
        help="The logging configuration file to load",
    )
    parser1.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Turn on all logging messages",
    )
    parser1.add_argument(
        "--debug",
        action="store_true",
        help="When an exception occurs, invoke the Python debugger, pdb",
    )
    parser1.add_argument(
        "--save-parameters",
        type=str,
        help="Save step parameters to specified file.",
    )
    parser1.add_argument(
        "--disable-crds-steppars",
        action="store_true",
        help="Disable retrieval of step parameter references files from CRDS",
    )
    known, _ = parser1.parse_known_args(args)

    try:
        if cls is None:
            step_class, config, name, config_file = _get_config_and_class(
                known.cfg_file_or_class[0]
            )
        else:
            config_file = known.config_file
            config = config_parser.load_config_file(config_file)
            step_class, name = Step._parse_class_and_name(
                config, config_file=config_file
            )
            step_class = cls

        log_config = None
        if known.verbose:
            if known.logcfg is not None:
                raise ValueError(
                    "If --verbose is set, a logging configuration file may not be"
                    " provided"
                )
            log_config = io.BytesIO(log.MAX_CONFIGURATION)
        elif known.logcfg is not None:
            if not os.path.exists(known.logcfg):
                raise OSError(f"Logging config {known.logcfg!r} not found")
            log_config = known.logcfg

        if log_config is not None:
            try:
                log.load_configuration(log_config)
            except Exception as e:
                raise ValueError(
                    f"Error parsing logging config {log_config!r}:\n{e}"
                ) from e
    except Exception as e:
        _print_important_message("ERROR PARSING CONFIGURATION:", str(e))
        parser1.print_help()
        raise

    debug_on_exception = known.debug

    # Determine whether CRDS should be queried for step parameters
    disable_crds_steppars = get_disable_crds_steppars(known.disable_crds_steppars)

    # This creates a config object from the spec file of the step class merged with
    # the spec files of the superclasses of the step class and adds arguments for
    # all of the expected reference files

    # load_spec_file is a method of both Step and Pipeline
    spec = step_class.load_spec_file()

    parser2 = _build_arg_parser_from_spec(spec, step_class, parent=parser1)

    args = parser2.parse_args(args)

    if cls is None:
        del args.cfg_file_or_class
    else:
        del args.config_file
    del args.logcfg
    del args.verbose
    del args.debug
    del args.save_parameters
    del args.disable_crds_steppars
    positional = args.args
    del args.args

    # This updates config (a ConfigObj) with the values from the command line arguments
    # Config is empty if class specified, otherwise contains values from config file
    # specified on command line
    _override_config_from_args(config, args)

    config = step_class.merge_config(config, config_file)

    if len(positional):
        input_file = positional[0]
        if args.input_dir:
            input_file = args.input_dir + "/" + input_file

        # Attempt to retrieve Step parameters from CRDS
        try:
            parameter_cfg = step_class.get_config_from_reference(
                input_file, disable=disable_crds_steppars
            )
        except (FileNotFoundError, OSError):
            log.log.warning(
                "Unable to open input file, cannot get parameters from CRDS"
            )
        else:
            if config:
                config_parser.merge_config(parameter_cfg, config)
            config = parameter_cfg
    else:
        log.log.info("No input file specified, unable to retrieve parameters from CRDS")

    # This is where the step is instantiated
    try:
        step = step_class.from_config_section(
            config,
            name=name,
            config_file=config_file,
        )
    except config_parser.ValidationError as e:
        # If the configobj validator failed, print usage information.
        _print_important_message("ERROR PARSING CONFIGURATION:", str(e))
        parser2.print_help()
        raise ValueError(str(e)) from e

    # Define the primary input file.
    # Always have an output_file set on the outermost step
    if len(positional):
        step.set_primary_input(positional[0])
        step.save_results = True

    log.log.info(f"Hostname: {os.uname()[1]}")
    log.log.info(f"OS: {os.uname()[0]}")

    # Save the step configuration
    if known.save_parameters:
        step.export_config(known.save_parameters, include_metadata=True)
        log.log.info(f"Step/Pipeline parameters saved to '{known.save_parameters}'")

    return step, step_class, positional, debug_on_exception


def step_from_cmdline(args, cls=None):
    """
    Create a step from a configuration file and run it.

    Parameters
    ----------
    args : list of str
        Commandline arguments

    cls : Step class
        The step class to use.  If none is provided, the step

    Returns
    -------
    step : Step instance
        If the config file has a `class` parameter, or the commandline
        specifies a class, the return value will be as instance of
        that class.

        Any parameters found in the config file or on the commandline
        will be set as member variables on the returned `Step`
        instance.
    """
    step, step_class, positional, debug_on_exception = just_the_step_from_cmdline(
        args, cls
    )

    try:
        profile_path = os.environ.pop("STPIPE_PROFILE", None)
        if profile_path:
            import cProfile

            cProfile.runctx("step.run(*positional)", globals(), locals(), profile_path)
        else:
            step.run(*positional)
    except Exception as e:
        _print_important_message(f"ERROR RUNNING STEP {step_class.__name__!r}:", str(e))

        if debug_on_exception:
            import pdb  # noqa: T100

            pdb.post_mortem()
        else:
            raise

    return step


def step_script(cls):
    import sys

    if not issubclass(cls, Step):
        raise AssertionError("cls must be a subclass of Step")

    return step_from_cmdline(sys.argv[1:], cls=cls)


def libpath(filepath):
    """Return the full path to the module library."""
    from os.path import abspath, dirname, join

    return join(dirname(abspath(__file__)), "lib", filepath)


__all__ = ['resample', 'resample_step',
'associations_lib_product_utils', 'associations_config',
'associations_lib_rules_elpp_base', 'associations_lib_member', 'associations_lib_diff',
'associations_lib_constraint', 'associations_pool', 'associations_lib_dms_base', 
'associations_association', 'associations_lib_utilities', 'associations_lib_process_list',
'associations_lib_ioregistry', 'associations_lib_keyvalue_registry', 
'associations_asn_from_list', 'associations_lib_acid', 'associations_lib_counter',
'associations_load_asn', 'associations_registry', 
'associations_lib_callback_registry',
'resample_gwcs_drizzle', 'assign_wcs_utils', 'resample_utils', 
'datamodels_library', 'associations_exceptions', 'lib_signal_slot',
'tweakreg_step', 'stpipe_utilities',
'stpipe_core', 'lib_suffix']


# import .resample
# import .associations_lib_rules_elpp_base, .associations_association, .associations_lib_constraint
# import .associations_pool, .associations_lib_dms_base, .associations_lib_member
# import .associations_lib_ioregistry, .associations_lib_keyvalue_registry, .associations_lib_utilities
# import .associations_asn_from_list, .associations_lib_acid, .associations_lib_counter
# import .associations_load_asn, .associations_registry, .associations_lib_product_utils
# import .associations_lib_callback_registry, .associations_lib_process_list
# import .assign_wcs_utils, .resample_gwcs_drizzle, .resample_utils
# import .associations_exceptions, .datamodels_library
# import .lib_signal_slot, .associations_config
# import .tweakreg_step, .stpipe_core, .lib_suffix

from . import resample, resample_step
from . import associations_lib_rules_elpp_base, associations_association, associations_lib_constraint
from . import associations_pool, associations_lib_dms_base, associations_lib_member
from . import associations_lib_ioregistry, associations_lib_keyvalue_registry, associations_lib_utilities
from . import associations_asn_from_list, associations_lib_acid, associations_lib_counter
from . import associations_load_asn, associations_registry, associations_lib_product_utils
from . import associations_lib_callback_registry, associations_lib_process_list
from . import assign_wcs_utils, resample_gwcs_drizzle, resample_utils
from . import associations_exceptions, datamodels_library
from . import lib_signal_slot, associations_config, stpipe_utilities
from . import tweakreg_step, stpipe_core, lib_suffix


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/resample/resample.py
"""


import logging
import os
import warnings
import json

import numpy as np
import psutil
from drizzle.resample import Drizzle
from spherical_geometry.polygon import SphericalPolygon

from stdatamodels.jwst import datamodels
from stdatamodels.jwst.library.basic_utils import bytes2human

import spike.jwstcal.resample_utils as resample_utils

from spike.jwstcal.associations_asn_from_list import asn_from_list

from spike.jwstcal.datamodels_library import ModelLibrary

from spike.jwstcal.model_blender_blender import ModelBlender


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["OutputTooLargeError", "ResampleData"]


class OutputTooLargeError(RuntimeError):
    """Raised when the output is too large for in-memory instantiation"""


class ResampleData:
    """
    This is the controlling routine for the resampling process.

    Notes
    -----
    This routine performs the following operations::

      1. Extracts parameter settings from input model, such as pixfrac,
         weight type, exposure time (if relevant), and kernel, and merges
         them with any user-provided values.
      2. Creates output WCS based on input images and define mapping function
         between all input arrays and the output array.
      3. Updates output data model with output arrays from drizzle, including
         a record of metadata from all input models.
    """

    def __init__(self, input_models, output=None, single=False, blendheaders=True,
                 pixfrac=1.0, kernel="square", fillval="NAN", wht_type="ivm",
                 good_bits=0, pscale_ratio=1.0, pscale=None, **kwargs):
        """
        Parameters
        ----------
        input_models : library of objects
            library of data models, one for each input image

        output : str
            filename for output

        kwargs : dict
            Other parameters.

            .. note::
                ``output_shape`` is in the ``x, y`` order.

            .. note::
                ``in_memory`` controls whether or not the resampled
                array from ``resample_many_to_many()``
                should be kept in memory or written out to disk and
                deleted from memory. Default value is `True` to keep
                all products in memory.
        """
        self.output_dir = None
        self.output_filename = output
        if output is not None and '.fits' not in str(output):
            self.output_dir = output
            self.output_filename = None
        self.intermediate_suffix = 'outlier_i2d'

        self.pscale_ratio = pscale_ratio
        self.single = single
        self.blendheaders = blendheaders
        self.pixfrac = pixfrac
        self.kernel = kernel
        self.fillval = fillval
        self.weight_type = wht_type
        self.good_bits = good_bits
        self.in_memory = kwargs.get('in_memory', True)
        self.input_pixscale0 = None  # computed pixel scale of the first image (deg)
        self._recalc_pscale_ratio = pscale is not None

        log.info(f"Driz parameter kernel: {self.kernel}")
        log.info(f"Driz parameter pixfrac: {self.pixfrac}")
        log.info(f"Driz parameter fillval: {self.fillval}")
        log.info(f"Driz parameter weight_type: {self.weight_type}")

        output_wcs = kwargs.get('output_wcs', None)
        output_shape = kwargs.get('output_shape', None)
        crpix = kwargs.get('crpix', None)
        crval = kwargs.get('crval', None)
        rotation = kwargs.get('rotation', None)

        self.asn_id = kwargs.get('asn_id', None)

        if pscale is not None:
            log.info(f'Output pixel scale: {pscale} arcsec.')
            pscale /= 3600.0
        else:
            log.info(f'Output pixel scale ratio: {pscale_ratio}')

        if output_wcs:
            # Use user-supplied reference WCS for the resampled image:
            self.output_wcs = output_wcs
            if output_shape is not None:
                self.output_wcs.array_shape = output_shape[::-1]

            if output_wcs.pixel_area is None:
                self.output_pix_area = compute_image_pixel_area(self.output_wcs)
                if self.output_pix_area is None:
                    raise ValueError(
                        "Unable to compute output pixel area from 'output_wcs'."
                    )
            else:
                self.output_pix_area = output_wcs.pixel_area

        else:
            # Define output WCS based on all inputs, including a reference WCS:
            self.output_wcs = resample_utils.make_output_wcs(
                input_models,
                ref_wcs=output_wcs,
                pscale_ratio=self.pscale_ratio,
                pscale=pscale,
                rotation=rotation,
                shape=None if output_shape is None else output_shape[::-1],
                crpix=crpix,
                crval=crval
            )
            # Estimate output pixel area in Sr. NOTE: in principle we could
            # use the same algorithm as for when output_wcs is provided by the
            # user.
            tr = self.output_wcs.pipeline[0].transform
            self.output_pix_area = (
                np.deg2rad(tr['cdelt1'].factor.value) *
                np.deg2rad(tr['cdelt2'].factor.value)
            )

        self.output_array_shape = tuple(self.output_wcs.array_shape)
        log.debug(f"Output mosaic size: {self.output_array_shape}")

        if pscale is None:
            pscale = np.rad2deg(np.sqrt(self.output_pix_area))
            log.info(f'Computed output pixel scale: {3600.0 * pscale} arcsec.')

        self.pscale = pscale  # in deg

        log.debug(f"Output mosaic size: {tuple(self.output_wcs.pixel_shape)}")

        allowed_memory = kwargs['allowed_memory']
        if allowed_memory is None:
            allowed_memory = os.environ.get('DMODEL_ALLOWED_MEMORY', allowed_memory)
        if allowed_memory:
            allowed_memory = float(allowed_memory)
            # make a small image model to get the dtype
            dtype = datamodels.ImageModel((1, 1)).data.dtype

            # get the available memory
            available_memory = psutil.virtual_memory().available + psutil.swap_memory().total

            # compute the output array size
            required_memory = np.prod(self.output_array_shape) * dtype.itemsize

            # compare used to available
            used_fraction = required_memory / available_memory
            can_allocate = used_fraction <= allowed_memory
        else:
            can_allocate = True

        if not can_allocate:
            raise OutputTooLargeError(
                f'Combined ImageModel size {self.output_array_shape} '
                f'requires {bytes2human(required_memory)}. '
                f'Model cannot be instantiated.'
            )

    def _create_output_model(self, ref_input_model=None):
        """ Create a new blank model and update it's meta with info from ``ref_input_model``. """
        output_model = datamodels.ImageModel(None)  # tuple(self.output_wcs.array_shape))

        # update meta data and wcs
        if ref_input_model is not None:
            output_model.update(ref_input_model)
        output_model.meta.wcs = self.output_wcs
        if self.output_pix_area is not None:
            output_model.meta.photometry.pixelarea_steradians = self.output_pix_area
            output_model.meta.photometry.pixelarea_arcsecsq = (
                self.output_pix_area * np.rad2deg(3600)**2
            )
        return output_model

    def do_drizzle(self, input_models):
        """Pick the correct drizzling mode based on self.single
        """
        if self.single:
            return self.resample_many_to_many(input_models)
        else:
            return self.resample_many_to_one(input_models)

    def _get_intensity_scale(self, img):
        """
        Compute an intensity scale from the input and output pixel area.

        For imaging data, the scaling is used to account for differences
        between the nominal pixel area and the average pixel area for
        the input data.

        For spectral data, the scaling is used to account for flux
        conservation with non-unity pixel scale ratios, when the
        data units are flux density.

        Parameters
        ----------
        img : DataModel
            The input data model.

        Returns
        -------
        iscale : float
            The scale to apply to the input data before drizzling.
        """
        input_pixflux_area = img.meta.photometry.pixelarea_steradians
        if input_pixflux_area:
            if 'SPECTRAL' in img.meta.wcs.output_frame.axes_type:
                # Use the nominal area as is
                input_pixel_area = input_pixflux_area

                # If input image is in flux density units, correct the
                # flux for the user-specified change to the spatial dimension
                if resample_utils.is_flux_density(img.meta.bunit_data):
                    input_pixel_area *= self.pscale_ratio
            else:
                img.meta.wcs.array_shape = img.data.shape
                input_pixel_area = compute_image_pixel_area(img.meta.wcs)
                if input_pixel_area is None:
                    raise ValueError(
                        "Unable to compute input pixel area from WCS of input "
                        f"image {repr(img.meta.filename)}."
                    )
                if self.input_pixscale0 is None:
                    self.input_pixscale0 = np.rad2deg(
                        np.sqrt(input_pixel_area)
                    )
                    if self._recalc_pscale_ratio:
                        self.pscale_ratio = self.pscale / self.input_pixscale0
            iscale = np.sqrt(input_pixflux_area / input_pixel_area)
        else:
            iscale = 1.0
        return iscale

    def resample_group(self, input_models, indices, compute_error=False):
        """Apply resample_many_to_many for one group

        Parameters
        ----------
        input_models : ModelLibrary

        indices : list

        compute_error : bool, optional
            If True, an approximate error image will be resampled
            alongside the science image.
        """
        output_model = None

        # Initialize the output with the wcs
        driz = Drizzle(
            out_shape=self.output_array_shape,
            kernel=self.kernel,
            fillval=self.fillval,
            disable_ctx=True,
        )
        # Also make a temporary model to hold error data
        if compute_error:
            driz_error = Drizzle(
                out_shape=self.output_array_shape,
                kernel=self.kernel,
                fillval=self.fillval,
                disable_ctx=True,
            )
        log.info(f"{len(indices)} exposures to drizzle together")
        for index in indices:
            img = input_models.borrow(index)
            if output_model is None:
                output_model = self._create_output_model(
                    ref_input_model=img
                )
                # Determine output file type from input exposure filenames
                # Use this for defining the output filename
                indx = img.meta.filename.rfind('.')
                output_type = img.meta.filename[indx:]
                output_root = '_'.join(img.meta.filename.replace(
                    output_type,
                    ''
                ).split('_')[:-1])
                output_model.meta.filename = (
                    f'{output_root}_'
                    f'{self.intermediate_suffix}{output_type}'
                )
                copy_asn_info_from_library(input_models, output_model)

            if isinstance(img, datamodels.SlitModel):
                # must call this explicitly to populate area extension
                # although the existence of this extension may not be necessary
                img.area = img.area

            iscale = self._get_intensity_scale(img)
            log.debug(f'Using intensity scale iscale={iscale}')

            inwht = resample_utils.build_driz_weight(
                img,
                weight_type=self.weight_type,
                good_bits=self.good_bits
            )

            # apply sky subtraction
            blevel = img.meta.background.level
            if not img.meta.background.subtracted and blevel is not None:
                data = img.data - blevel
            else:
                data = img.data

            xmin, xmax, ymin, ymax = resample_utils._resample_range(
                data.shape,
                img.meta.wcs.bounding_box
            )
            pixmap = resample_utils.calc_gwcs_pixmap(
                img.meta.wcs,
                self.output_wcs,
                img.data.shape,
            )

            driz.add_image(
                data=data,
                exptime=img.meta.exposure.exposure_time,  # GWCSDrizzle.add_image param default was 1.0
                pixmap=pixmap,
                scale=iscale,
                weight_map=inwht,
                wht_scale=1.0,  # hard-coded for JWST count-rate data
                pixfrac=self.pixfrac,
                in_units="cps",  # GWCSDrizzle.add_image param default
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
            )
            del data

            # make an approximate error image by drizzling it
            # in the same way the image is handled
            if compute_error:
                driz_error.add_image(
                    data=img.err,
                    exptime=img.meta.exposure.exposure_time,  # GWCSDrizzle.add_image param default
                    pixmap=pixmap,
                    scale=iscale,
                    weight_map=inwht,
                    wht_scale=1.0,  # hard-coded for JWST count-rate data
                    pixfrac=self.pixfrac,
                    in_units="cps",  # GWCSDrizzle.add_image param default
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )
            input_models.shelve(img, index, modify=False)
            del img

        output_model.data = driz.out_img
        output_model.wht = driz.out_wht
        del driz
        # copy the drizzled error into the output model
        if compute_error:
            output_model.err = driz_error.out_img
            del driz_error

        return output_model

    def resample_many_to_many(self, input_models):
        """Resample many inputs to many outputs where outputs have a common frame.

        Coadd only different detectors of the same exposure, i.e. map NRCA5 and
        NRCB5 onto the same output image, as they image different areas of the
        sky.

        Used for outlier detection
        """
        output_models = []
        for group_id, indices in input_models.group_indices.items():

            output_model = self.resample_group(input_models, indices)

            if not self.in_memory:
                # Write out model to disk, then return filename
                output_name = output_model.meta.filename
                if self.output_dir is not None:
                    output_name = os.path.join(self.output_dir, output_name)
                output_model.save(output_name)
                log.info(f"Saved model in {output_name}")
                output_models.append(output_name)
            else:
                output_models.append(output_model)

        if self.in_memory:
            # build ModelLibrary as a list of in-memory models
            return ModelLibrary(output_models, on_disk=False)
        else:
            # build ModelLibrary as an association from the output files
            # this saves memory if there are multiple groups
            asn = asn_from_list(output_models, product_name='outlier_i2d')
            asn_dict = json.loads(asn.dump()[1])  # serializes the asn and converts to dict
            return ModelLibrary(asn_dict, on_disk=True)

    def resample_many_to_one(self, input_models):
        """Resample and coadd many inputs to a single output.

        Used for stage 3 resampling
        """
        output_model = None

        if self.blendheaders:
            blender = ModelBlender(
                blend_ignore_attrs=[
                    'meta.photometry.pixelarea_steradians',
                    'meta.photometry.pixelarea_arcsecsq',
                    'meta.filename',
                ]
            )

        driz = Drizzle(
            out_shape=self.output_array_shape,
            kernel=self.kernel,
            fillval=self.fillval,
            max_ctx_id=len(input_models),
            disable_ctx=False,
        )
        self._init_variance_arrays()
        self._init_exptime_counters()

        log.info("Resampling science and variance data")

        leading_group_idx = [v[0] for v in input_models.group_indices.values()]

        with input_models:
            for idx, img in enumerate(input_models):
                if output_model is None:
                    output_model = self._create_output_model(
                        ref_input_model=img
                    )
                    # Determine output file type from input exposure filenames
                    # Use this for defining the output filename
                    output_model.meta.filename = self.output_filename
                    output_model.meta.resample.weight_type = self.weight_type
                    output_model.meta.resample.pointings = len(input_models.group_names)

                    # copy over asn information
                    copy_asn_info_from_library(input_models, output_model)

                if idx in leading_group_idx:
                    self._update_exptime(img)

                if self.blendheaders:
                    blender.accumulate(img)

                iscale = self._get_intensity_scale(img)
                log.debug(f'Using intensity scale iscale={iscale}')

                inwht = resample_utils.build_driz_weight(
                    img,
                    weight_type=self.weight_type,
                    good_bits=self.good_bits,
                )

                # apply sky subtraction
                blevel = img.meta.background.level
                if not img.meta.background.subtracted and blevel is not None:
                    data = img.data - blevel
                else:
                    data = img.data.copy()

                in_image_limits = resample_utils._resample_range(
                    data.shape,
                    img.meta.wcs.bounding_box
                )
                xmin, xmax, ymin, ymax = in_image_limits

                pixmap = resample_utils.calc_gwcs_pixmap(
                    img.meta.wcs,
                    output_model.meta.wcs,
                    data.shape,
                )

                driz.add_image(
                    data=data,
                    exptime=img.meta.exposure.exposure_time,  # GWCSDrizzle.add_image param default
                    pixmap=pixmap,
                    scale=iscale,
                    weight_map=inwht,
                    wht_scale=1.0,  # hard-coded for JWST count-rate data
                    pixfrac=self.pixfrac,
                    in_units="cps",  # GWCSDrizzle.add_image param default
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )
                # Resample variance arrays in input_models to output_model
                self._resample_variance_arrays(
                    model=img,
                    iscale=iscale,
                    inwht=inwht,
                    pixmap=pixmap,
                    in_image_limits=in_image_limits,
                    output_shape=self.output_array_shape,
                )

                del data, inwht

                input_models.shelve(img)

        # Since the context array is dynamic, it must be re-assigned
        # back to the product's `con` attribute.
        output_model.data = driz.out_img
        output_model.wht = driz.out_wht
        if driz.out_ctx is not None:
            output_model.con = driz.out_ctx

        del driz

        # compute final variances:
        self._compute_resample_variance_totals(output_model)

        var_components = [
            output_model.var_rnoise,
            output_model.var_poisson,
            output_model.var_flat
        ]
        output_model.err = np.sqrt(np.nansum(var_components, axis=0))

        # nansum returns zero for input that is all NaN -
        # set those values to NaN instead
        all_nan = np.all(np.isnan(var_components), axis=0)
        output_model.err[all_nan] = np.nan

        if self.blendheaders:
            blender.finalize_model(output_model)

        self._get_exptime_totals(output_model)

        return ModelLibrary([output_model,], on_disk=False)

    def _init_variance_arrays(self):
        shape = self.output_array_shape
        self._weighted_rn_var = np.full(shape, np.nan, dtype=np.float32)
        self._weighted_pn_var = np.full(shape, np.nan, dtype=np.float32)
        self._weighted_flat_var = np.full(shape, np.nan, dtype=np.float32)
        self._total_weight_rn_var = np.zeros(shape, dtype=np.float32)
        self._total_weight_pn_var = np.zeros(shape, dtype=np.float32)
        self._total_weight_flat_var = np.zeros(shape, dtype=np.float32)

    def _resample_variance_arrays(self, model, iscale, inwht, pixmap,
                                  in_image_limits, output_shape):
        xmin, xmax, ymin, ymax = in_image_limits

        # Do the read noise variance first, so it can be
        # used for weights if needed
        rn_var = self._resample_one_variance_array(
            "var_rnoise",
            input_model=model,
            iscale=iscale,
            inwht=inwht,
            pixmap=pixmap,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
        )

        # Find valid weighting values in the variance
        if rn_var is not None:
            mask = (rn_var > 0) & np.isfinite(rn_var)
        else:
            mask = np.full_like(rn_var, False)

        # Set the weight for the image from the weight type
        weight = np.ones(output_shape)
        if self.weight_type == "ivm" and rn_var is not None:
            weight[mask] = rn_var[mask] ** -1
        elif self.weight_type == "exptime":
            if resample_utils.check_for_tmeasure(model):
                weight[:] = model.meta.exposure.measurement_time
            else:
                weight[:] = model.meta.exposure.exposure_time

        # Weight and add the readnoise variance
        # Note: floating point overflow is an issue if variance weights
        # are used - it can't be squared before multiplication
        if rn_var is not None:
            mask = (rn_var >= 0) & np.isfinite(rn_var) & (weight > 0)
            self._weighted_rn_var[mask] = np.nansum(
                [
                    self._weighted_rn_var[mask],
                    rn_var[mask] * weight[mask] * weight[mask]
                ],
                axis=0
            )
            self._total_weight_rn_var[mask] += weight[mask]

        # Now do poisson and flat variance, updating only valid new values
        # (zero is a valid value; negative, inf, or NaN are not)
        pn_var = self._resample_one_variance_array(
            "var_poisson",
            input_model=model,
            iscale=iscale,
            inwht=inwht,
            pixmap=pixmap,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
        )
        if pn_var is not None:
            mask = (pn_var >= 0) & np.isfinite(pn_var) & (weight > 0)
            self._weighted_pn_var[mask] = np.nansum(
                [
                    self._weighted_pn_var[mask],
                    pn_var[mask] * weight[mask] * weight[mask]
                ],
                axis=0
            )
            self._total_weight_pn_var[mask] += weight[mask]

        flat_var = self._resample_one_variance_array(
            "var_flat",
            input_model=model,
            iscale=iscale,
            inwht=inwht,
            pixmap=pixmap,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
        )
        if flat_var is not None:
            mask = (flat_var >= 0) & np.isfinite(flat_var) & (weight > 0)
            self._weighted_flat_var[mask] = np.nansum(
                [
                    self._weighted_flat_var[mask],
                    flat_var[mask] * weight[mask] * weight[mask]
                ],
                axis=0
            )
            self._total_weight_flat_var[mask] += weight[mask]

    def _compute_resample_variance_totals(self, output_model):
        # Divide by the total weights, squared, and set in the output model.
        # Zero weight and missing values are NaN in the output.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "invalid value*", RuntimeWarning)
            warnings.filterwarnings("ignore", "divide by zero*", RuntimeWarning)

            output_variance = (
                self._weighted_rn_var / self._total_weight_rn_var /
                self._total_weight_rn_var
            )
            setattr(output_model, "var_rnoise", output_variance)

            output_variance = (
                self._weighted_pn_var / self._total_weight_pn_var /
                self._total_weight_pn_var
            )
            setattr(output_model, "var_poisson", output_variance)

            output_variance = (
                self._weighted_flat_var / self._total_weight_flat_var /
                self._total_weight_flat_var
            )
            setattr(output_model, "var_flat", output_variance)

        del (
            self._weighted_rn_var,
            self._weighted_pn_var,
            self._weighted_flat_var,
            self._total_weight_rn_var,
            self._total_weight_pn_var,
            self._total_weight_flat_var,
        )

    def _resample_one_variance_array(self, name, input_model, iscale,
                                     inwht, pixmap,
                                     xmin=None, xmax=None, ymin=None, ymax=None):
        """Resample one variance image from an input model.

        The error image is passed to drizzle instead of the variance, to
        better match kernel overlap and user weights to the data, in the
        pixel averaging process. The drizzled error image is squared before
        returning.
        """
        variance = getattr(input_model, name)
        if variance is None or variance.size == 0:
            log.debug(
                f"No data for '{name}' for model "
                f"{repr(input_model.meta.filename)}. Skipping ..."
            )
            return

        elif variance.shape != input_model.data.shape:
            log.warning(
                f"Data shape mismatch for '{name}' for model "
                f"{repr(input_model.meta.filename)}. Skipping ..."
            )
            return

        output_shape = self.output_array_shape

        # Resample the error array. Fill "unpopulated" pixels with NaNs.
        driz = Drizzle(
            out_shape=output_shape,
            kernel=self.kernel,
            fillval=np.nan,
            disable_ctx=True
        )

        log.debug(f"Pixmap shape: {pixmap[:,:,0].shape}")
        log.debug(f"Input Sci shape: {variance.shape}")
        log.debug(f"Output Sci shape: {output_shape}")

        # Call 'drizzle' to perform image combination
        log.info(f"Drizzling {variance.shape} --> {output_shape}")

        driz.add_image(
            data=np.sqrt(variance),
            exptime=input_model.meta.exposure.exposure_time,
            pixmap=pixmap,
            scale=iscale,
            weight_map=inwht,
            wht_scale=1.0,  # hard-coded for JWST count-rate data
            pixfrac=self.pixfrac,
            in_units="cps",
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
        )

        return driz.out_img ** 2

    def _init_exptime_counters(self):
        self._total_exposure_time = 0.
        self._exposure_times = {'start': [], 'end': []}
        self._duration = 0.0
        self._total_measurement_time = 0.0
        self._measurement_time_failures = []

    def _update_exptime(self, model):
        self._total_exposure_time += model.meta.exposure.exposure_time
        if not resample_utils.check_for_tmeasure(model):
            self._measurement_time_failures.append(1)
        else:
            self._total_measurement_time += model.meta.exposure.measurement_time
            self._measurement_time_failures.append(0)
        self._exposure_times['start'].append(model.meta.exposure.start_time)
        self._exposure_times['end'].append(model.meta.exposure.end_time)
        self._duration += model.meta.exposure.duration

    def _get_exptime_totals(self, output_model):
        # Update some basic exposure time values based on output_model
        output_model.meta.exposure.exposure_time = self._total_exposure_time
        if not any(self._measurement_time_failures):
            output_model.meta.exposure.measurement_time = self._total_measurement_time
        output_model.meta.exposure.start_time = min(self._exposure_times['start'])
        output_model.meta.exposure.end_time = max(self._exposure_times['end'])

        # Update other exposure time keywords:
        # XPOSURE (identical to the total effective exposure time, EFFEXPTM)
        xposure = self._total_exposure_time
        output_model.meta.exposure.effective_exposure_time = xposure
        # DURATION (identical to TELAPSE, elapsed time)
        output_model.meta.exposure.duration = self._duration
        output_model.meta.exposure.elapsed_exposure_time = self._duration

        del (
            self._total_exposure_time,
            self._exposure_times,
            self._duration,
            self._total_measurement_time,
            self._measurement_time_failures,
        )

    def update_exposure_times(self, output_model, input_models):
        """Modify exposure time metadata in-place"""
        self._init_exptime_counters()
        with input_models:
            for _, indices in input_models.group_indices.items():
                model = input_models.borrow(indices[0])
                self._update_exptime(model)
        self._get_exptime_totals(output_model)


def _get_boundary_points(xmin, xmax, ymin, ymax, dx=None, dy=None, shrink=0):
    """
    xmin, xmax, ymin, ymax - integer coordinates of pixel boundaries
    step - distance between points along an edge
    shrink - number of pixels by which to reduce `shape`

    Returns a list of points and the area of the rectangle
    """
    nx = xmax - xmin + 1
    ny = ymax - ymin + 1

    if dx is None:
        dx = nx
    if dy is None:
        dy = ny

    if nx - 2 * shrink < 1 or ny - 2 * shrink < 1:
        raise ValueError("Image size is too small.")

    sx = max(1, int(np.ceil(nx / dx)))
    sy = max(1, int(np.ceil(ny / dy)))

    xmin += shrink
    xmax -= shrink
    ymin += shrink
    ymax -= shrink

    size = 2 * sx + 2 * sy
    x = np.empty(size)
    y = np.empty(size)

    b = np.s_[0:sx]  # bottom edge
    r = np.s_[sx:sx + sy]  # right edge
    t = np.s_[sx + sy:2 * sx + sy]  # top edge
    l = np.s_[2 * sx + sy:2 * sx + 2 * sy]  # left

    x[b] = np.linspace(xmin, xmax, sx, False)
    y[b] = ymin
    x[r] = xmax
    y[r] = np.linspace(ymin, ymax, sy, False)
    x[t] = np.linspace(xmax, xmin, sx, False)
    y[t] = ymax
    x[l] = xmin
    y[l] = np.linspace(ymax, ymin, sy, False)

    area = (xmax - xmin) * (ymax - ymin)
    center = (0.5 * (xmin + xmax), 0.5 * (ymin + ymax))

    return x, y, area, center, b, r, t, l


def compute_image_pixel_area(wcs):
    """ Computes pixel area in steradians.
    """
    if wcs.array_shape is None:
        raise ValueError("WCS must have array_shape attribute set.")

    valid_polygon = False
    spatial_idx = np.where(np.array(wcs.output_frame.axes_type) == 'SPATIAL')[0]

    ny, nx = wcs.array_shape
    ((xmin, xmax), (ymin, ymax)) = wcs.bounding_box

    xmin = max(0, int(xmin + 0.5))
    xmax = min(nx - 1, int(xmax - 0.5))
    ymin = max(0, int(ymin + 0.5))
    ymax = min(ny - 1, int(ymax - 0.5))
    if xmin > xmax:
        (xmin, xmax) = (xmax, xmin)
    if ymin > ymax:
        (ymin, ymax) = (ymax, ymin)

    k = 0
    dxy = [1, -1, -1, 1]
    ra, dec, center = np.nan, np.nan, (np.nan, np.nan)
    while xmin < xmax and ymin < ymax:
        try:
            x, y, image_area, center, b, r, t, l = _get_boundary_points(
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
                dx=min((xmax - xmin) // 4, 15),
                dy=min((ymax - ymin) // 4, 15)
            )
        except ValueError:
            return None

        world = wcs(x, y)
        ra = world[spatial_idx[0]]
        dec = world[spatial_idx[1]]

        limits = [ymin, xmax, ymax, xmin]

        for j in range(4):
            sl = [b, r, t, l][k]
            if not (np.all(np.isfinite(ra[sl])) and
                    np.all(np.isfinite(dec[sl]))):
                limits[k] += dxy[k]
                k = (k + 1) % 4
                break
            k = (k + 1) % 4
        else:
            valid_polygon = True
            break

        ymin, xmax, ymax, xmin = limits

    if not valid_polygon:
        return None

    world = wcs(*center)
    wcenter = (world[spatial_idx[0]], world[spatial_idx[1]])

    sky_area = SphericalPolygon.from_radec(ra, dec, center=wcenter).area()
    if sky_area > 2 * np.pi:
        log.warning(
            "Unexpectedly large computed sky area for an image. "
            "Setting area to: 4*Pi - area"
        )
        sky_area = 4 * np.pi - sky_area
    pix_area = sky_area / image_area

    return pix_area


def copy_asn_info_from_library(library, output_model):
    """
    Transfer association information from the input library to the output model.

    Parameters
    ----------
    library : ModelLibrary
        The input library of data models.

    output_model : DataModel
        The output data model to which the association information will be copied.
    """
    if not hasattr(library, "asn"):
        # No ASN table, occurs when input comes from ModelContainer in spectroscopic modes
        # in this case do nothing; the asn info will be passed along later
        # by code inside ResampleSpecStep
        return
    if (asn_pool := library.asn.get("asn_pool", None)) is not None:
        output_model.meta.asn.pool_name = asn_pool
    if (
        asn_table_name := library.asn.get("table_name", None)
    ) is not None:
        output_model.meta.asn.table_name = asn_table_name


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/lib/prune.py
"""


import logging
from collections import defaultdict

import spike.jwstcal.associations_lib_diff as diff
from spike.jwstcal.associations_lib_product_utils import get_product_names, sort_by_candidate
import spike.jwstcal.associations_config as config

__all__ = ['prune']

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Duplicate association counter
# Used in function `prune_remove`
DupCount = 0

def prune(asns):
    """Remove duplicates and subset associations

    Situations where extraneous associations can occur are:

    - duplicate memberships
    - duplicate product names

    Associations with different product names but same memberships arise when
    different levels of candidates gather the same membership, such as
    OBSERVATION vs. GROUP. Associations of the lower level candidate are preferred.

    Associations with the same product name can occur in Level 2 when both an OBSERVATION
    candidate and a BACKGROUND candidate associations are created. The association that is
    a superset of members is the one chosen.

    Parameters
    ----------
    asns : [Association[,...]]
        Associations to prune

    Returns
    -------
    pruned : [Association[,...]]
        Pruned list of associations
    """
    pruned = prune_duplicate_associations(asns)
    pruned = prune_duplicate_products(pruned)
    return pruned

def prune_duplicate_associations(asns):
    """Remove duplicate associations in favor of lower level versions

    Main use case: For Level 3 associations, multiple associations with the
    same membership, but different levels, can be created. Remove duplicate
    associations of higher level.

    The assumption is that there is only one product per association, before
    merging.

    Parameters
    ----------
    asns : [Association[,...]]
        Associations to prune

    Returns
    -------
    pruned : [Association[,...]]
        Pruned list of associations

    """
    known_dups, valid_asns = identify_dups(asns)

    ordered_asns = sort_by_candidate(valid_asns)
    pruned = list()
    while True:
        try:
            original = ordered_asns.pop()
        except IndexError:
            break
        pruned.append(original)
        to_prune = list()
        for asn in ordered_asns:
            try:
                diff.compare_product_membership(original['products'][0], asn['products'][0])
            except AssertionError:
                continue
            to_prune.append(asn)
        prune_remove(ordered_asns, to_prune, known_dups)

    return pruned + known_dups


def prune_duplicate_products(asns):
    """Remove duplicate products in favor of higher level versions

    The assumption is that there is only one product per association, before
    merging

    Parameters
    ----------
    asns: [Association[,...]]
        Associations to prune

    Returns
    pruned: [Association[,...]]
        Pruned list of associations

    """
    known_dups, valid_asns = identify_dups(asns)

    product_names, dups = get_product_names(valid_asns)
    if not dups:
        return asns

    ordered_asns = sort_by_candidate(valid_asns)
    asn_by_product = defaultdict(list)
    for asn in ordered_asns:
        asn_by_product[asn['products'][0]['name']].append(asn)

    full_prune = list()
    for product in dups:
        dup_asns = asn_by_product[product]
        to_keep = dup_asns.copy()
        for asn in dup_asns:
            # Association has been removed, ignore
            if asn in full_prune:
                continue

            # Check against the set of associations to be kept.
            to_prune = list()
            for entrant in to_keep:
                if entrant == asn:
                    continue

                # Check for differences. If none, then the associations are exact duplicates.
                try:
                    diff.compare_product_membership(asn['products'][0], entrant['products'][0])
                except diff.MultiDiffError as diffs:
                    # If one is a pure subset, remove the smaller association.
                    if len(diffs) == 1 and isinstance(diffs[0], diff.SubsetError):
                        if len(entrant['products'][0]['members']) > len(asn['products'][0]['members']):
                            asn, entrant = entrant, asn
                        to_prune.append(entrant)
                        continue

                    # If the difference is only in suffix, this is an acceptable duplication of product names.
                    # Trap and do not report.
                    try:
                        diff.compare_product_membership(asn['products'][0], entrant['products'][0], strict_expname=False)
                    except diff.MultiDiffError:
                        # Something is different. Report but do not remove.
                        logger.warning('Following associations have the same product name but significant differences.')
                        logger.warning('Association 1: %s', asn)
                        logger.warning('Association 2: %s', entrant)
                        logger.warning('Diffs: %s', diffs)

                else:
                    # Associations are exactly the same. Discard the logically lesser one.
                    # Due to the sorting, this should be the current `asn`
                    to_prune.append(entrant)

            # Update lists.
            full_prune.extend(to_prune)
            for asn in to_prune:
                to_keep.remove(asn)

    prune_remove(ordered_asns, full_prune, known_dups)
    return ordered_asns + known_dups


def prune_remove(remove_from, to_remove, known_dups):
    """Remove or rename associations to be pruned

    Default behavior is to remove associations listed in the `to_remove`
    list from the `remove_from` list.

    However, if `config.DEBUG` is `True`, that association is simply
    renamed, adding the string "dupXXXXX" as a prefix to the association's
    name.

    Parameters
    ----------
    remove_from : [Association[,...]]
        The list of associations from which associations will be removed.
        List is modified in-place.

    to_remove : [Association[,...]]
        The list of associations to remove from the `remove_from` list.

    known_dups : [Association[,...]]
        Known duplicates. New ones are added by this function
        if debugging is in effect.
    """
    global DupCount

    if to_remove:
        logger.debug('Duplicate associations found: %s', to_remove)
    for asn in to_remove:
        remove_from.remove(asn)
        if config.DEBUG:
            DupCount += 1
            asn.asn_name = f'dup{DupCount:05d}_{asn.asn_name}'
            known_dups.append(asn)


def identify_dups(asns):
    """Separate associations based on whether they have already been identified as dups

    Parameters
    ----------
    asns: [Association[,...]]
        Associations to prune

    Returns
    identified, valid : [Association[,...]], [Association[,...]]
        Dup-identified and valid associations
    """
    identified = list()
    valid = list()
    for asn in asns:
        if asn.asn_name.startswith('dup'):
            identified.append(asn)
        else:
            valid.append(asn)
    return identified, valid